from typing import List, Optional, Union

from graia.application.message.elements.internal import At, Image


class Command(object):
    '''指令相关

    `cmd` 为指令，
    `args` 为参数，
    `argsList` 为参数列表，
    '''
    cmd: Optional[str]
    args: Optional[str]
    argsList: Optional[List[str]]

    #TODO 迁移指令解析逻辑


class AtList(List[At]):
    '''
    At 的增强版，允许你更方便地操作 At 列表
    AtList 本身也是一个列表，因此可以使用大部分的 list 特性
    '''
    ints: List[int]

    @staticmethod
    def __getTarget(at: Union[At, int]) -> Optional[int]:
        if isinstance(at, At):
            return at.target
        elif isinstance(at, int):
            return at
        else:
            raise TypeError('Must be Union[At, int]')

    def __init__(self, atList: List[At]) -> None:
        self.ints = [i.target for i in atList]
        super().__init__(atList)

    def __contains__(self, at: Union[At, int]) -> bool:
        return self.__getTarget(at) in self.ints

    def append(self, at: Union[At, int]) -> None:
        if isinstance(at, At):
            atObj = at
        elif isinstance(at, int):
            atObj =  At(at)
        else:
            raise TypeError('Must be Union[At, int]')
        super().append(atObj)

    def __set(self, l: List[At]) -> None:
        '''
        内部函数，用于更新 AtList
        '''
        self.__init__(l)

    def get(self, at: Union[At, int]) -> At:
        '''
        根据 QQ 号或 `At` 对象获取 `At` 对象本身
        '''
        target = self.__getTarget(at)
        ats: List[At] = self.copy()
        for i in ats:
            if i.target == target:
                return i
        else:
            return None

    def __del(self, at: Union[At, int]) -> None:
        '''
        内部函数，用于删除 `AtList` 中的某一项
        '''
        target = self.__getTarget(at)
        ats = self.copy()
        for i in range(len(ats)):
            if ats[i].target == target:
                del ats[i]
                self.__set(ats)
                break

    def delete(self, at: Union[At, int]) -> None:
        '''
        根据传入的 QQ 号或 `At` 对象删除 `AtList` 中的所有匹配项
        '''
        while self.__contains__(at):
            self.__del(at)


class ImagePro(Image):
    async def save2file(self, filepath: str) -> None:
        '''
        将图片保存至文件

        Args:
            filepath: 文件路径
        '''
        imageBytes = await self.http_to_bytes()
        with open(filepath, 'wb') as f:
            f.write(imageBytes)



from typing import List
from .utils import createFiles

# 文件不存在则创建
adminsList = 'admins.list'
blockusersList = 'blockusers.list'
createFiles([adminsList, blockusersList])

class Permission(object):
    '''权限处理'''
    adminsList = adminsList
    blockusersList = blockusersList

    qq: int

    def __init__(self, qq: int) -> None:
        '''检查用户权限或进行权限操作

        :params qq: QQ 号'''
        self.qq = qq

    def _readBlockList(self) -> List[int]:
        with open(self.blockusersList, 'r+') as f:
            return [int(_i) for _i in f.read().split()]

    def _writeBlockList(self, l: list) -> None:
        _l = [str(_i) for _i in l]
        with open(self.blockusersList, 'w+') as f:
            f.writelines([f'{_i}\n' for _i in _l])

    def _readAdminList(self) -> List[int]:
        with open(self.adminsList, 'r+') as f:
            _l = [int(_i) for _i in f.read().split()]
        return _l

    def isAdmin(self) -> bool:
        '''是否为 admin'''
        return self.qq in self._readAdminList()

    def isBlocked(self) -> bool:
        '''是否已被 block'''
        _l = self._readBlockList()
        return self.qq in _l

    def blockme(self) -> bool:
        '''封禁自己

        成功 True

        :params userId: 被封禁对象的 QQ'''
        _l = self._readBlockList()
        if self.qq in _l:  # 用户已被 block
            return False
        else:
            _l.append(self.qq)
            self._writeBlockList(_l)
            return True

    def unblockme(self) -> bool:
        '''解除自己的封禁

        成功 True

        :params userId: 被解除封禁的对象的 QQ'''
        _l = self._readBlockList()
        if not self.qq in _l:  # 用户没被 block
            return False
        else:
            del _l[_l.index(self.qq)]
            self._writeBlockList(_l)
            return True



from typing import List, Optional, Tuple, Union

from graia.application.entry import (At, GroupMessage, MessageChain, Plain,
                                     Quote, Source)
from graia.application.event.messages import FriendMessage
from graia.application.friend import Friend
from graia.application.group import Member
from graia.application.message.elements.internal import Image

from .elements import AtList, ImagePro
from .permission import Permission


class NEM(object):
    '''
    Not Enough Messages -- 更好的消息解析器

    示例：
    ``` python
    @bcc.receiver(GroupMessage)
    async def group_message_listener(app: GraiaMiraiApplication, group: Group, _gm: GroupMessage):
        nem = NEM(_gm)
        await app.sendGroupMessage(group, MessageChain.create([Plain(nem.plain_message)]))
    ```

    属性解释：（带 `*` 号的表示此项有可能为 `None`）
    - `nem.permission` 为消息发送者的 `Permission` 对象
    - `nem.chain` 为原消息链（`MessageChain`）
    - `nem.sender` 为消息发送者的 `Member` 或 `Friend` 实例
    - `nem.sender_id` 为消息发送者的 QQ 号
    - `nem.source` 为原消息链的 `Source`，可用于回复
    - `nem.at` *为消息中所有的 `At` 列表，使用了 `AtList` 拥有特殊功能
    - `nem.images` *为 `ImagePro` 列表
    - `nem.plain_message` *为所有的 `Plain` 拼接而成的字符串
    - `nem.plain_message_source` *为被回复的消息中所有的 `Plain` 拼接而成的字符串
    '''
    _commandSymbol: str
    chain: MessageChain
    sender: Union[Member, Friend]
    sender_id: int
    plain_message: Optional[str]
    plain_message_source: Optional[str]
    source: Source
    at: AtList
    images: List[ImagePro]
    permission: Permission

    class Command(object):
        '''指令相关

        `cmd` 为指令，
        `args` 为参数，
        `argsList` 为参数列表，
        '''
        cmd: Optional[str]
        args: Optional[str]
        argsList: Optional[List[str]]

    def __init__(self, _message: Union[GroupMessage, FriendMessage], _command_symbol: str = '&') -> None:
        '''
        初始化 NEM 解析器

        Args:
            _message: `Message` 对象
            _command_symbol: 指令标识符，默认为 `'&'`
        '''
        self._commandSymbol = self._commandSymbol or _command_symbol
        self.chain = _message.messageChain
        self.sender = _message.sender
        self.sender_id = _message.sender.id
        self.plain_message = self._getPlainMessage(self.chain)
        self.at = self._getAt()
        self.Command.cmd, self.Command.args, self.Command.argsList = self._getCommand()
        self.plain_message_source = self._getQuotePlainMessage()
        self.source = self._getSource()
        self.images = self._makeImageProList()
        self.permission = Permission(self.sender_id)

    @staticmethod
    def _getPlainMessage(_messagechain: MessageChain) -> Optional[str]:
        if Plain in _messagechain:
            _text = str()
            for _i in _messagechain[Plain]:
                i_text: str = _i.text
                if not i_text.strip().isspace():
                    _text = f'{_text} {i_text}'
            return _text.strip()
        else:
            return None

    def _getAt(self) -> AtList:
        atList = self.chain[At]
        return AtList(atList)

    def _getQuotePlainMessage(self) -> Optional[str]:
        if Quote in self.chain:
            _quote_message: Quote = self.chain[Quote][0]
            _origin_messagechain: MessageChain = _quote_message.origin
            _plain_message = self._getPlainMessage(_origin_messagechain)
            return _plain_message
        else:
            return None

    def _getSource(self) -> Source:
        return self.chain[Source][0]

    def makeMessageBody(self, plain_message: str) -> str:
        '''
        帮助你实现更好的指令解析，适用于一些消息转发机器人的特殊处理

        继承 `NEM` 类并重写这个函数以实现此功能
        '''
        return plain_message  # 不作操作，原样返回

    def _getCommand(self) -> Tuple[Optional[str], Optional[str], Optional[List[str]]]:
        if not self.plain_message:
            return None, None, None
        if self._commandSymbol in self.plain_message:
            _message_body: str = self.makeMessageBody(self.plain_message)
            _splited_message: list = _message_body.split(' ', 1)
            _command: str = _splited_message[0]
            _args: Optional[str] = _splited_message[1] if len(
                _splited_message) > 1 else None
            _argsList: Optional[List[str]] = _args.split() if _args else None
            cmd: str = _command.replace(
                self._commandSymbol, '')  # 只留下命令主体而不考虑命令标识符
            return cmd, _args, _argsList
        else:
            return None, None, None

    def _makeImageProList(self) -> Optional[List[ImagePro]]:
        if not Image in self.chain:
            return None
        images = self.chain[Image]
        return [ImagePro(i) for i in images]



from typing import List, TypeVar
import os.path

from graia.application.event.messages import FriendMessage, GroupMessage, TempMessage

AnyMessage = TypeVar('AnyMessage', GroupMessage, FriendMessage, TempMessage)

def createFiles(files: List[str]) -> None:
    for file in files:
        if not os.path.exists(file):
            with open(file, 'w') as f:
                f.write('')


import setuptools

with open('README.md', 'r', encoding='utf8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='graiax-nem',
    version='0.0.4.dev3',
    author='Xiao_Jin',
    keywords='graia graiax nem graiax-nem graia-message',
    description='适用于 Graia Framework 的消息解析过滤器',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/jinzhijie/graiax-nem',
    packages=['graiax/nem'],
    install_requires=['graia-application-mirai'],
    python_requires='>=3.7',
)



'''
 _   _   _____   __  __ 
| \ | | | ____| |  \/  |
|  \| | |  _|   | |\/| |
| |\  | | |___  | |  | |
|_| \_| |_____| |_|  |_|

  Not Enough Messages
    更好的消息解析器
需配合 Graia Framework 使用


Usage:
``` python
from graiaX.nem import NEM

...

@bcc.receiver(GroupMessage)
async def group_message_listener(app: GraiaMiraiApplication, group: Group, _gm: GroupMessage):
    nem = NEM(_gm)
    await app.sendGroupMessage(group, MessageChain.create([Plain(nem.plain_message)]))

...
```
'''

from .nem import NEM



import re
from typing import Callable, List

from graia.application.entry import Group, GroupMessage
from graia.application.event.messages import FriendMessage
from graia.broadcast.exceptions import ExecutionStop
from graia.broadcast.builtin.decoraters import Depend

from .nem import NEM


class FrontFilters(object):
    '''过滤器的前置修饰器，无法直接使用'''
    @staticmethod
    def requirePlain(func: Callable[[GroupMessage], Callable]):
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if not nem.plain_message:
                raise ExecutionStop()
            return func(gm)
        return wrapper

    @staticmethod
    def requireNotBanded(func: Callable[[GroupMessage], Callable]):
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if nem.permission.isBlocked():
                raise ExecutionStop()
            return func(gm)
        return wrapper

    @staticmethod
    def depend(func: Callable) -> Depend:
        def wrapper(*args, **kwargs):
            return Depend(func(*args, **kwargs))
        return wrapper


class GroupFilters(object):
    '''
    Group 消息过滤器，作为 `headless_decoraters` 使用

    请查看源码以了解所有可用的过滤器
    '''

    # GroupMessage
    @staticmethod
    @FrontFilters.depend
    def exceptGroups(group_list: List[int]) -> Callable:
        '''在指定群聊内禁用'''
        def wrapper(group: Group):
            if group.id in group_list:
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def inGroups(group_list: List[int]) -> Depend:
        '''仅在指定群聊内启用'''
        def wrapper(group: Group):
            if not group.id in group_list:
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def adminOnly() -> Depend:
        '''仅管理员'''
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if not nem.permission.isAdmin():
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def onCommand(command_name: str) -> Depend:
        '''在特定指令时执行

        Args:
            command_name: 指令名称（不带标识符）'''
        @FrontFilters.requirePlain
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if nem.Command.cmd != command_name:
                raise ExecutionStop()
        return wrapper
    
    @staticmethod
    @FrontFilters.depend
    def onAt(qq: int) -> Depend:
        '''当 `qq` 在群内被 At 时执行

        Args:
            qq: QQ 号'''
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if not qq in nem.at:
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def onWord(word: str) -> Depend:
        '''在消息中有特定文字时执行

        Args:
            word: 文字'''
        @FrontFilters.requirePlain
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if not word in nem.plain_message:
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def onWords(words_list: List[str]) -> Depend:
        '''在消息中有特定文字时执行

        Args:
            words_list: 关键字列表'''
        @FrontFilters.requirePlain
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            inList = False
            for word in words_list:
                if word in nem.plain_message:
                    inList = True
                    break
            if not inList:
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def onMatch(pattern: str) -> Depend:
        '''在消息匹配正则表达式时执行

        Args:
            pattern: 正则表达式（必须为 str）'''
        @FrontFilters.requirePlain
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            if not re.match(pattern, nem.plain_message):
                raise ExecutionStop()
        return wrapper

    @staticmethod
    @FrontFilters.depend
    def onMatchs(pattern_list: List[str]) -> Depend:
        '''在消息匹配正则表达式列表中的任何一项时执行

        Args:
            pattern_list: 正则表达式列表'''
        @FrontFilters.requirePlain
        @FrontFilters.requireNotBanded
        def wrapper(gm: GroupMessage):
            nem = NEM(gm)
            matched: bool = False
            for p in pattern_list:
                if re.match(p, nem.plain_message):
                    matched = True
                    break
            if not matched:
                raise ExecutionStop()
        return wrapper


class FriendFilters(object):
    @staticmethod
    @FrontFilters.depend
    def onFriend(qq: int) -> Depend:
        def wrapper(fm: FriendMessage):
            if fm.sender.id != qq:
                raise ExecutionStop()
        return wrapper
